@@ -13,13 +13,13 @@ from functools import cmp_to_key
 from itertools import zip_longest
 from typing import Any, Optional
 
-from elementpath._typing import Callable, Iterable, Iterator
+from collections.abc import Callable, Iterable, Iterator
 from elementpath.protocols import ElementProtocol
 from elementpath.exceptions import xpath_error
 from elementpath.datatypes import UntypedAtomic, AnyURI, AbstractQName
 from elementpath.collations import UNICODE_CODEPOINT_COLLATION, CollationManager
-from elementpath.xpath_nodes import XPathNode, ElementNode, AttributeNode, \
-    NamespaceNode, TextNode, CommentNode, ProcessingInstructionNode, DocumentNode
+from elementpath.xpath_nodes import XPathNode, EtreeElementNode, TextAttributeNode, \
+    NamespaceNode, TextNode, CommentNode, ProcessingInstructionNode, EtreeDocumentNode
 from elementpath.xpath_tokens import XPathToken, XPathFunction, XPathMap, XPathArray
 
 
@@ -28,7 +28,7 @@ def deep_equal(seq1: Iterable[Any],
                collation: Optional[str] = None,
                token: Optional[XPathToken] = None) -> bool:
 
-    etree_node_types = (ElementNode, CommentNode, ProcessingInstructionNode)
+    etree_node_types = (EtreeElementNode, CommentNode, ProcessingInstructionNode)
 
     def etree_deep_equal(e1: ElementProtocol, e2: ElementProtocol) -> bool:
         if cm.ne(e1.tag, e2.tag):
@@ -66,37 +66,44 @@ def deep_equal(seq1: Iterable[Any],
 
             if (value1 is None) ^ (value2 is None):
                 return False
+            elif isinstance(value1, (XPathArray, XPathMap, XPathNode)) ^ \
+                    isinstance(value2, (XPathArray, XPathMap, XPathNode)):
+                return False
             elif value1 is None:
                 return True
-            elif isinstance(value1, XPathNode) ^ isinstance(value2, XPathNode):
-                return False
+            elif isinstance(value1, XPathMap):
+                assert isinstance(value2, XPathMap)
+                return value1 == value2
+            elif isinstance(value1, XPathArray):
+                assert isinstance(value2, XPathArray)
+                return value1 == value2
             elif isinstance(value1, XPathNode):
                 assert isinstance(value2, XPathNode)
-                if value1.kind != value2.kind:
+                if value1.__class__ != value2.__class__:
                     return False
                 elif isinstance(value1, etree_node_types):
                     assert isinstance(value2, etree_node_types)
-                    if not etree_deep_equal(value1.elem, value2.elem):
+                    if not etree_deep_equal(value1.obj, value2.obj):
                         return False
-                elif isinstance(value1, DocumentNode):
-                    assert isinstance(value2, DocumentNode)
+                elif isinstance(value1, EtreeDocumentNode):
+                    assert isinstance(value2, EtreeDocumentNode)
                     for child1, child2 in zip_longest(value1, value2):
                         if child1 is None or child2 is None:
                             return False
-                        elif child1.kind != child2.kind:
+                        elif child1.__class__ != child2.__class__:
                             return False
                         elif isinstance(child1, etree_node_types):
                             assert isinstance(child2, etree_node_types)
-                            if not etree_deep_equal(child1.elem, child2.elem):
+                            if not etree_deep_equal(child1.obj, child2.obj):
                                 return False
                         elif isinstance(child1, TextNode):
                             assert isinstance(child2, TextNode)
-                            if cm.ne(child1.value, child2.value):
+                            if cm.ne(child1.obj, child2.obj):
                                 return False
 
-                elif cm.ne(value1.value, value2.value):
+                elif cm.ne(value1.obj, value2.obj):
                     return False
-                elif isinstance(value1, AttributeNode):
+                elif isinstance(value1, TextAttributeNode):
                     if cm.ne(value1.name, value2.name):
                         return False
                 elif isinstance(value1, NamespaceNode):
@@ -174,7 +181,7 @@ def deep_compare(obj1: Any,
                  token: Optional[XPathToken] = None) -> int:
 
     msg_tmpl = "Sorting failed, cannot compare {!r} with {!r}"
-    etree_node_types = (ElementNode, CommentNode, ProcessingInstructionNode)
+    etree_node_types = (EtreeElementNode, CommentNode, ProcessingInstructionNode)
     result: int = 0
 
     def iter_object(obj: Any) -> Iterator[Any]:
@@ -187,9 +194,10 @@ def deep_compare(obj1: Any,
 
     def etree_deep_compare(e1: ElementProtocol, e2: ElementProtocol) -> int:
         nonlocal result
-        result = cm.strcoll(e1.tag, e2.tag)
-        if result:
-            return result
+        if not callable(e1.tag) and not callable(e2.tag):
+            result = cm.strcoll(e1.tag, e2.tag)
+            if result:
+                return result
 
         result = cm.strcoll((e1.text or '').strip(), (e2.text or '').strip())
         if result:
@@ -244,51 +252,51 @@ def deep_compare(obj1: Any,
                 raise xpath_error('XPTY0004', msg, token=token)
             elif isinstance(value1, XPathNode):
                 assert isinstance(value2, XPathNode)
-                if value1.kind != value2.kind:
+                if value1.__class__ != value2.__class__:
                     msg = f"cannot compare {type(value1)} with {type(value2)}"
                     raise xpath_error('XPTY0004', msg, token=token)
                 elif isinstance(value1, etree_node_types):
                     assert isinstance(value2, etree_node_types)
-                    result = etree_deep_compare(value1.elem, value2.elem)
+                    result = etree_deep_compare(value1.obj, value2.obj)
                     if result:
                         return result
-                elif isinstance(value1, DocumentNode):
-                    assert isinstance(value2, DocumentNode)
+                elif isinstance(value1, EtreeDocumentNode):
+                    assert isinstance(value2, EtreeDocumentNode)
                     for child1, child2 in zip_longest(value1, value2):
                         if child1 is None:
                             return -1
                         elif child2 is None:
                             return 1
-                        elif child1.kind != child2.kind:
+                        elif child1.__class__ != child2.__class__:
                             msg = f"cannot compare {type(child1)} with {type(child2)}"
                             raise xpath_error('XPTY0004', msg, token=token)
                         elif isinstance(child1, etree_node_types):
                             assert isinstance(child2, etree_node_types)
-                            result = etree_deep_compare(child1.elem, child2.elem)
+                            result = etree_deep_compare(child1.obj, child2.obj)
                             if result:
                                 return result
                         elif isinstance(child1, TextNode):
                             assert isinstance(child2, TextNode)
                             result = cm.strcoll(
-                                child1.value.strip(), child2.value.strip()
+                                child1.obj.strip(), child2.obj.strip()
                             )
                             if result:
                                 return result
-                else:
-                    result = cm.strcoll(value1.value, value2.value)
+                elif isinstance(value1, TextNode):
+                    assert isinstance(value2, TextNode)
+                    result = cm.strcoll(value1.obj, value2.obj)
+                    if result:
+                        return result
+                elif isinstance(value1, TextAttributeNode):
+                    assert isinstance(value2, TextAttributeNode)
+                    result = cm.strcoll(value1.name or '', value2.name or '')
+                    if result:
+                        return result
+                elif isinstance(value1, NamespaceNode):
+                    assert isinstance(value2, NamespaceNode)
+                    result = cm.strcoll(value1.prefix or '', value2.prefix or '')
                     if result:
                         return result
-
-                    if isinstance(value1, AttributeNode):
-                        assert isinstance(value2, AttributeNode)
-                        result = cm.strcoll(value1.name or '', value2.name or '')
-                        if result:
-                            return result
-                    elif isinstance(value1, NamespaceNode):
-                        assert isinstance(value2, NamespaceNode)
-                        result = cm.strcoll(value1.prefix or '', value2.prefix or '')
-                        if result:
-                            return result
             else:
                 try:
                     if isinstance(value1, bool):
@@ -315,7 +323,8 @@ def deep_compare(obj1: Any,
 
                     elif isinstance(value1, float):
                         if math.isnan(value1):
-                            if not math.isnan(value2):
+                            if not isinstance(value2, (float, Decimal)) \
+                                    or not math.isnan(value2):
                                 return -1
                         elif math.isinf(value1):
                             if value1 != value2:
