@@ -16,13 +16,13 @@ import time
 import re
 import os.path
 import unicodedata
+from collections.abc import Iterator
 from copy import copy
 from decimal import Decimal, DecimalException
 from string import ascii_letters
-from typing import cast, List, Optional, Type, Union
+from typing import cast, Optional, Union
 from urllib.parse import urlsplit, quote as urllib_quote
 
-from elementpath._typing import Iterator
 from elementpath.aliases import Emptiable, AnyNsmapType
 from elementpath.exceptions import ElementPathValueError
 from elementpath.helpers import Patterns, is_idrefs, is_xml_codepoint, round_number
@@ -34,7 +34,7 @@ from elementpath.namespaces import XML_NAMESPACE, get_namespace, split_expanded_
 from elementpath.compare import deep_equal
 from elementpath.sequence_types import match_sequence_type
 from elementpath.xpath_context import ContextType, ItemType, XPathSchemaContext
-from elementpath.xpath_nodes import XPathNode, DocumentNode, ElementNode, SchemaElementNode
+from elementpath.xpath_nodes import XPathNode, DocumentNode, ElementNode, EtreeElementNode
 from elementpath.xpath_tokens import XPathFunction
 from elementpath.regex import RegexError, translate_pattern
 from elementpath.collations import CollationManager
@@ -152,8 +152,10 @@ def evaluate_namespace_uri_for_prefix_function(
     elem = self.get_argument(context, index=1)
     if not isinstance(elem, ElementNode):
         raise self.error('FORG0006', '2nd argument %r is not an element node' % elem)
+    if not isinstance(elem, EtreeElementNode):
+        return []
 
-    ns_uris = {get_namespace(e.tag) for e in elem.elem.iter() if not callable(e.tag)}
+    ns_uris = {get_namespace(e.tag) for e in elem.obj.iter() if not callable(e.tag)}
     for p, uri in self.parser.namespaces.items():
         if uri in ns_uris:
             if p == prefix:
@@ -178,7 +180,7 @@ def select_in_scope_prefixes_function(self: XPathFunction, context: ContextType
     if not isinstance(arg, ElementNode):
         raise self.error('XPTY0004', 'argument %r is not an element node' % arg)
 
-    elem = arg.elem
+    elem = arg.obj
     if isinstance(context, XPathSchemaContext):
         # For schema context returns prefixes of static namespaces
         for pfx, uri in self.parser.namespaces.items():
@@ -270,8 +272,13 @@ def evaluate_node_name_function(self: XPathFunction, context: ContextType = None
     elif name.startswith('{'):
         # name is a QName in extended format
         namespace, local_name = split_expanded_name(name)
+        if not namespace:
+            return QName('', local_name)
+
         for pfx, uri in self.parser.namespaces.items():
             if uri == namespace:
+                if not pfx:
+                    return QName(uri, local_name)
                 return QName(uri, '{}:{}'.format(pfx, local_name))
         raise self.error('FONS0004', 'no prefix found for namespace {}'.format(namespace))
     else:
@@ -409,7 +416,7 @@ def evaluate_avg_function(self: XPathFunction, context: ContextType = None) \
     if self.context is not None:
         context = self.context
 
-    values: List[AtomicType] = []
+    values: list[AtomicType] = []
     for item in self[0].atomization(context):
         if isinstance(item, UntypedAtomic):
             values.append(self.cast_to_double(item.value))
@@ -431,10 +438,10 @@ def evaluate_avg_function(self: XPathFunction, context: ContextType = None) \
                 return []
             raise self.error('FORG0006', err)
     elif all(isinstance(x, int) for x in values):
-        result = sum(cast(List[int], values)) / Decimal(len(values))
+        result = sum(cast(list[int], values)) / Decimal(len(values))
         return int(result) if result % 1 == 0 else result
     elif all(isinstance(x, (int, Decimal)) for x in values):
-        return sum(cast(List[Decimal], values)) / Decimal(len(values))
+        return sum(cast(list[Decimal], values)) / Decimal(len(values))
     elif all(not isinstance(x, DoubleProxy) for x in values):
         try:
             return sum(
@@ -468,25 +475,25 @@ def evaluate_max_min_functions(self: XPathFunction, context: ContextType = None)
         elif all(isinstance(x, str) for x in values):
             if to_any_uri:
                 return AnyURI(aggregate_func(
-                    cast(List[str], values)
+                    cast(list[str], values)
                 ))
         elif any(isinstance(x, str) for x in values):
             if any(isinstance(x, ArithmeticProxy) for x in values):
                 raise self.error('FORG0006', "cannot compare strings with numeric data")
         elif all(isinstance(x, (Decimal, int)) for x in values):
             return aggregate_func(
-                cast(List[str], values)
+                cast(list[str], values)
             )
         elif any(isinstance(x, float) and math.isnan(x) for x in values):
             return float_class('NaN')
         elif all(isinstance(x, (int, float, Decimal)) for x in values):
             return float_class(
-                aggregate_func(cast(List[NumericType], values))
+                aggregate_func(cast(list[NumericType], values))
             )
         return aggregate_func(values)  # type: ignore[type-var]
 
-    values: List[AtomicType] = []
-    float_class: Union[Type[Float10], Type[float]] = Float10
+    values: list[AtomicType] = []
+    float_class: Union[type[Float10], type[float]] = Float10
     to_any_uri = None
     aggregate_func = max if self.symbol == 'max' else min
 
@@ -565,7 +572,7 @@ def select_distinct_values_function(self: XPathFunction, context: ContextType =
 
     def distinct_values(case_insensitive: bool = False) -> Iterator[AtomicType]:
         nan = False
-        results: List[AtomicType] = []
+        results: list[AtomicType] = []
         for value in self[0].atomization(context):
             if case_insensitive and isinstance(value, (str, bytes)):
                 value = value.casefold()
@@ -850,7 +857,7 @@ def evaluate_replace_function(self: XPathFunction, context: ContextType = None)
 @method(function('tokenize', nargs=(1, 3),
                  sequence_types=('xs:string?', 'xs:string', 'xs:string', 'xs:string*')))
 def evaluate_tokenize_function(self: XPathFunction, context: ContextType = None) \
-        -> Emptiable[Union[List[str], str]]:
+        -> Emptiable[Union[list[str], str]]:
     if self.context is not None:
         context = self.context
 
@@ -962,7 +969,7 @@ def evaluate_codepoints_to_string_function(
 @method(function('string-to-codepoints', nargs=1,
                  sequence_types=('xs:string?', 'xs:integer*')))
 def evaluate_string_to_codepoints_function(self: XPathFunction, context: ContextType = None) \
-        -> List[int]:
+        -> list[int]:
     arg = self.get_argument(self.context or context, cls=str)
     return [ord(c) for c in arg] if arg else []
 
@@ -1476,34 +1483,34 @@ def evaluate_lang_function(self: XPathFunction, context: ContextType = None) ->
 
     if not isinstance(item, ElementNode):
         raise self.error('XPTY0004')
-    elif isinstance(item, SchemaElementNode):
-        return False
+    elif isinstance(item, EtreeElementNode):
+        try:
+            attr = item.obj.attrib[XML_LANG]
+        except KeyError:
+            if len(self) > 1 or context is None:
+                return False
 
-    try:
-        attr = item.elem.attrib[XML_LANG]
-    except KeyError:
-        if len(self) > 1 or context is None:
-            return False
-
-        for elem in context.iter_ancestors():
-            if isinstance(elem, ElementNode):
-                if XML_LANG in elem.elem.attrib:
-                    lang = cast(str, elem.elem.attrib[XML_LANG])
-                    break
+            for elem in context.iter_ancestors():
+                if isinstance(elem, EtreeElementNode):
+                    if XML_LANG in elem.obj.attrib:
+                        lang = cast(str, elem.obj.attrib[XML_LANG])
+                        break
+            else:
+                return False
         else:
-            return False
+            if not isinstance(attr, str):
+                return False
+            lang = attr.strip()
+
+        test_lang: str = self.get_argument(context, cls=str)
+        if test_lang is None:
+            test_lang = ''
+
+        test_lang = test_lang.strip().lower()
+        lang = lang.strip().lower()
+        return lang == test_lang or lang.startswith(test_lang) and lang[len(test_lang)] == '-'
     else:
-        if not isinstance(attr, str):
-            return False
-        lang = attr.strip()
-
-    test_lang: str = self.get_argument(context, cls=str)
-    if test_lang is None:
-        test_lang = ''
-
-    test_lang = test_lang.strip().lower()
-    lang = lang.strip().lower()
-    return lang == test_lang or lang.startswith(test_lang) and lang[len(test_lang)] == '-'
+        return False
 
 
 ###
@@ -1540,55 +1547,26 @@ def select_id_function(self: XPathFunction, context: ContextType = None) -> Iter
     if root is None:
         return
 
-    # TODO: PSVI bindings with also xsi:type evaluation
     for element in root.iter_descendants():
-        if not isinstance(element, ElementNode):
+        if not isinstance(element, EtreeElementNode):
             continue
 
-        if element.elem.text in idrefs:
-            if self.parser.schema is not None:
-                xsd_element = self.parser.schema.find(element.path, self.parser.namespaces)
-                if xsd_element is None or not hasattr(xsd_element, 'type') or \
-                        xsd_element.type is None or not xsd_element.type.is_key():
-                    continue
-
-            idrefs.remove(element.elem.text)
+        if element.obj.text in idrefs and element.is_id:
+            idrefs.remove(element.obj.text)
             if self.symbol == 'id':
                 yield element
             else:
                 parent = element.parent
-                if isinstance(parent, ElementNode):
+                if isinstance(parent, EtreeElementNode):
                     yield parent
-            continue  # pragma: no cover
+        else:
+            for attr in element.attributes:
+                if not isinstance(attr.obj, str):
+                    continue
 
-        for attr in element.attributes:
-            if not isinstance(attr.value, str):
-                continue
-
-            if attr.value in idrefs:
-                if attr.name == XML_ID:
-                    idrefs.remove(attr.value)
+                if attr.obj in idrefs and attr.is_id:
+                    idrefs.remove(attr.obj)
                     yield element
-                    break
-
-                if self.parser.schema is None:
-                    continue
-
-                xsd_element = self.parser.schema.find(element.path, self.parser.namespaces)
-                if xsd_element is None:
-                    continue
-
-                try:
-                    xsd_attribute = xsd_element.attrib[attr.name]
-                except KeyError:
-                    continue
-                else:
-                    if xsd_attribute.type is None or not xsd_attribute.type.is_key():
-                        continue  # pragma: no cover
-
-                    idrefs.remove(attr.value)
-                    yield element
-                    break
 
 
 @method(function('idref', nargs=(1, 2), sequence_types=('xs:string*', 'node()', 'node()*')))
@@ -1605,12 +1583,12 @@ def select_idref_function(self: XPathFunction, context: ContextType = None) \
         return
     elif not isinstance(node, XPathNode):
         raise self.error('XPTY0004')
-    elif isinstance(node, (ElementNode, DocumentNode)):
+    elif isinstance(node, (EtreeElementNode, DocumentNode)):
         for element in node.iter_descendants():
-            if not isinstance(element, ElementNode):
+            if not isinstance(element, EtreeElementNode):
                 continue
 
-            text = element.elem.text
+            text = element.obj.text
             if text and is_idrefs(text) and \
                     any(v in text.split() for x in ids for v in x.split()):
                 yield element
@@ -1618,8 +1596,8 @@ def select_idref_function(self: XPathFunction, context: ContextType = None) \
 
             if element.attributes:
                 for attr in element.attributes:  # pragma: no cover
-                    if attr.name != XML_ID and isinstance(attr.value, str) and \
-                            any(v in attr.value.split() for x in ids for v in x.split()):
+                    if attr.name != XML_ID and isinstance(attr.obj, str) and \
+                            any(v in attr.obj.split() for x in ids for v in x.split()):
                         yield element
                         break
 
@@ -1679,7 +1657,7 @@ def evaluate_doc_functions(self: XPathFunction, context: ContextType = None) \
         sequence_type = 'document-node()'
 
     if not match_sequence_type(doc, sequence_type, self.parser):
-        msg = f"Type does not match sequence type {sequence_type!r}"
+        msg = f"type does not match sequence type {sequence_type!r}"
         raise self.error('XPDY0050', msg)
 
     return doc if self.symbol == 'doc' else True
@@ -1687,7 +1665,7 @@ def evaluate_doc_functions(self: XPathFunction, context: ContextType = None) \
 
 @method(function('collection', nargs=(0, 1), sequence_types=('xs:string?', 'node()*')))
 def evaluate_collection_function(self: XPathFunction, context: ContextType = None) \
-        -> Emptiable[List[XPathNode]]:
+        -> Emptiable[list[XPathNode]]:
     if self.context is not None:
         context = self.context
 
@@ -1717,7 +1695,7 @@ def evaluate_collection_function(self: XPathFunction, context: ContextType = Non
             return collection
 
     if not match_sequence_type(collection, sequence_type, self.parser):
-        msg = f"Type does not match sequence type {sequence_type!r}"
+        msg = f"type does not match sequence type {sequence_type!r}"
         raise self.error('XPDY0050', msg)
 
     return collection
