@@ -10,35 +10,33 @@
 """
 XPath 3.0 implementation - part 3 (functions)
 """
-import sys
 import decimal
 import os
 import re
 import codecs
 import math
+import zoneinfo
+from collections.abc import Iterator
 from copy import copy
 from itertools import zip_longest
-from typing import cast, Any, Dict, List, Optional, Tuple, Union, Type, Set
+from typing import cast, Any, Optional, Union
 from urllib.parse import urlsplit
 from urllib.request import urlopen
 from urllib.error import URLError
 
-from elementpath._typing import Iterator
 from elementpath.aliases import Emptiable
-from elementpath.protocols import ElementProtocol
 from elementpath.exceptions import ElementPathError
 from elementpath.tdop import MultiLabel
 from elementpath.helpers import OCCURRENCE_INDICATORS, Patterns, \
     is_xml_codepoint, node_position
 from elementpath.namespaces import get_expanded_name, split_expanded_name, \
     XPATH_FUNCTIONS_NAMESPACE, XSD_NAMESPACE
-from elementpath.datatypes import xsd10_atomic_types, NumericProxy, QName, Date10, \
-    DateTime10, Time, AnyURI, UntypedAtomic, AtomicType, NumericType, NMToken, \
-    Idref, Entity
+from elementpath.datatypes import NumericProxy, QName, Date10, DateTime10, Time, \
+    AnyURI, UntypedAtomic, AtomicType, NumericType, NMToken, Idref, Entity
 from elementpath.sequence_types import is_sequence_type, match_sequence_type
-from elementpath.etree import defuse_xml, etree_iter_paths
-from elementpath.xpath_nodes import XPathNode, ElementNode, TextNode, AttributeNode, \
-    NamespaceNode, DocumentNode, ProcessingInstructionNode, CommentNode
+from elementpath.etree import defuse_xml
+from elementpath.xpath_nodes import XPathNode, ElementNode, NamespaceNode, \
+    DocumentNode, EtreeElementNode, SchemaElementNode
 from elementpath.tree_builders import get_node_tree
 from elementpath.xpath_tokens import XPathToken, ValueToken, XPathFunction, XPathConstructor
 from elementpath.serialization import get_serialization_params, serialize_to_xml, \
@@ -53,11 +51,6 @@ from .xpath30_helpers import UNICODE_DIGIT_PATTERN, DECIMAL_DIGIT_PATTERN, \
     format_digits, int_to_words, parse_datetime_picture, parse_datetime_marker, \
     ordinal_suffix
 
-if sys.version_info < (3, 9):
-    zoneinfo = None
-else:
-    import zoneinfo
-
 FORMAT_INTEGER_TOKENS = {'A', 'a', 'i', 'I', 'w', 'W', 'Ww'}
 
 DECL_PARAM_PATTERN = re.compile(r'([^\d\W][\w.\-\u00B7\u0300-\u036F\u203F\u2040]*)\s*=\s*')
@@ -81,10 +74,10 @@ class _InlineFunction(XPathFunction):
     body: Optional[XPathToken] = None
     "Body of anonymous inline function."
 
-    variables: Optional[Dict[str, Any]] = None
+    variables: Optional[dict[str, Any]] = None
     "In-scope variables linked by let and for expressions and arguments."
 
-    varnames: Optional[List[str]] = None
+    varnames: Optional[list[str]] = None
     "Inline function arguments varnames."
 
     def __str__(self) -> str:
@@ -172,7 +165,7 @@ class _InlineFunction(XPathFunction):
                     elif context.root is not None:
                         context.item = context.root
 
-                args = cast(Tuple[FunctionArgType], (context.item,))
+                args = cast(tuple[FunctionArgType], (context.item,))
 
             partial_function = False
             if self.variables is None:
@@ -293,7 +286,7 @@ class _InlineFunction(XPathFunction):
 
         return self
 
-    def evaluate(self, context: ContextType = None) -> Union[ItemType, List[ItemType]]:
+    def evaluate(self, context: ContextType = None) -> Union[ItemType, list[ItemType]]:
         if context is None:
             raise self.missing_context()
         elif self.label.endswith('function'):
@@ -316,7 +309,7 @@ class _InlineFunction(XPathFunction):
         nargs = len([tk and not tk for tk in self._items if tk.symbol == '?'])
         assert nargs, "a partial function requires at least a placeholder token"
 
-        self._name = None
+        self._name = None  # noqa
         self.label = 'inline partial function'
         self.nargs = nargs
 
@@ -666,7 +659,7 @@ def evaluate_format_number_function(self: XPathFunction, context: ContextType =
 
     minus_sign = decimal_format['minus-sign']
 
-    prefix = ''
+    prefix: str = ''
     if value >= 0:
         subpic = sub_pictures[0]
     else:
@@ -689,7 +682,8 @@ def evaluate_format_number_function(self: XPathFunction, context: ContextType =
         suffix = percent_sign
         subpic = subpic[:-1]
 
-        if value.as_tuple().exponent < 0:
+        exponent = value.as_tuple().exponent
+        if isinstance(exponent, int) and exponent < 0:
             value *= 100
         else:
             value = decimal.Decimal(int(value) * 100)
@@ -698,7 +692,8 @@ def evaluate_format_number_function(self: XPathFunction, context: ContextType =
         suffix = per_mille_sign
         subpic = subpic[:-1]
 
-        if value.as_tuple().exponent < 0:
+        exponent = value.as_tuple().exponent
+        if isinstance(exponent, int) and exponent < 0:
             value *= 1000
         else:
             value = decimal.Decimal(int(value) * 1000)
@@ -843,7 +838,7 @@ function('format-time', nargs=(2, 5),
 @method('format-time')
 def evaluate_format_date_time_functions(self: XPathFunction, context: ContextType = None) \
         -> Emptiable[str]:
-    cls: Type[Union[DateTime10, Date10, Time]]
+    cls: type[Union[DateTime10, Date10, Time]]
     if self.symbol == 'format-dateTime':
         cls = DateTime10
         invalid_markers = ''
@@ -937,7 +932,7 @@ def evaluate_format_date_time_functions(self: XPathFunction, context: ContextTyp
                  sequence_types=('xs:string?', 'xs:string', 'xs:string',
                                  'element(fn:analyze-string-result)')))
 def evaluate_analyze_string_function(self: XPathFunction, context: ContextType = None) \
-        -> ElementProtocol:
+        -> ElementNode:
     if self.context is not None:
         context = self.context
 
@@ -1061,7 +1056,7 @@ def evaluate_analyze_string_function(self: XPathFunction, context: ContextType =
     else:
         root = context.etree.XML(''.join(lines))
 
-    return cast(ElementProtocol, get_node_tree(root=root, namespaces=self.parser.namespaces))
+    return cast(ElementNode, get_node_tree(root=root, namespaces=self.parser.namespaces))
 
 
 ###
@@ -1073,65 +1068,24 @@ def evaluate_path_function(self: XPathFunction, context: ContextType = None) ->
     elif context is None:
         raise self.missing_context()
 
-    if isinstance(context, XPathSchemaContext):
-        return []
-    elif not self:
+    if not self:
         item = context.item
     else:
         item = self.get_argument(context)
         if item is None:
             return []
 
-    suffix = ''
-    if isinstance(item, DocumentNode):
-        return '/'
-    elif isinstance(item, (ElementNode, CommentNode, ProcessingInstructionNode)):
-        elem = item.elem
-    elif not isinstance(item, XPathNode) or item.parent is None:
+    if not isinstance(item, XPathNode):
         return []
-    elif isinstance(item, TextNode):
-        elem = item.parent.elem
-        suffix = '/text()[1]'
-    elif isinstance(item, AttributeNode) and item.name is not None:
-        elem = item.parent.elem
-        if item.name.startswith('{'):
-            suffix = f'/@Q{item.name}'
-        else:
-            suffix = f'/@{item.name}'
-    elif isinstance(item, NamespaceNode):
-        elem = item.parent.elem
-        if item.prefix:
-            suffix = f'/namespace::{item.prefix}'
-        else:
-            suffix = f'/namespace::*[Q{{{XPATH_FUNCTIONS_NAMESPACE}}}local-name()=""]'
+    elif context.root is None or (root_node := item.root_node) is not context.root:
+        # The context has no root or the root is not the root of the item node
+        return []
+    elif not isinstance(root_node, (DocumentNode, SchemaElementNode)):
+        # It's a fragment: add fn:root() to select the root position
+        path = item.path[len(root_node.path):]
+        return f"Q{{{XPATH_FUNCTIONS_NAMESPACE}}}root(){path}"
     else:
-        return []
-
-    if isinstance(context.root, DocumentNode):
-        root = context.root.getroot().elem
-        if root.tag.startswith('{'):
-            path = f'/Q{root.tag}[1]'
-        else:
-            path = f'/Q{{}}{root.tag}[1]'
-    elif context.root is None:
-        return []
-    else:
-        # If root is an element use the function that returns the root of the tree
-        root = context.root.elem
-        path = 'Q{%s}root()' % XPATH_FUNCTIONS_NAMESPACE
-
-    if isinstance(item, ProcessingInstructionNode):
-        if item.parent is None or isinstance(item.parent, DocumentNode):
-            return f'/processing-instruction({item.name})[{context.position}]'
-    elif isinstance(item, CommentNode):
-        if item.parent is None or isinstance(item.parent, DocumentNode):
-            return f'/comment()[{context.position}]'
-
-    for e, path in etree_iter_paths(root, path):
-        if e is elem:
-            return path + suffix
-    else:
-        return []
+        return item.path or []
 
 
 @method(function('has-children', nargs=(0, 1), sequence_types=('node()?', 'xs:boolean')))
@@ -1153,7 +1107,7 @@ def evaluate_has_children_function(self: XPathFunction, context: ContextType = N
             raise self.error('XPTY0004', 'argument must be a node')
 
     return isinstance(item, DocumentNode) or \
-        isinstance(item, ElementNode) and (len(item.elem) > 0 or item.elem.text is not None)
+        isinstance(item, EtreeElementNode) and (len(item.obj) > 0 or item.obj.text is not None)
 
 
 @method(function('innermost', nargs=1, sequence_types=('node()*', 'node()*')))
@@ -1169,7 +1123,7 @@ def select_innermost_function(self: XPathFunction, context: ContextType = None)
 
     ancestors = {x for context.item in nodes for x in context.iter_ancestors(axis='ancestor')}
     results = {x for x in nodes if x not in ancestors}
-    yield from cast(List[XPathNode], sorted(results, key=node_position))
+    yield from cast(list[XPathNode], sorted(results, key=node_position))
 
 
 @method(function('outermost', nargs=1, sequence_types=('node()*', 'node()*')))
@@ -1180,7 +1134,7 @@ def select_outermost_function(self: XPathFunction, context: ContextType = None)
 
     context = copy(context)
 
-    nodes: Union[List[ItemType], Set[ItemType]]
+    nodes: Union[list[ItemType], set[ItemType]]
     nodes = [e for e in self[0].select(context)]
     if any(not isinstance(x, XPathNode) for x in nodes):
         raise self.error('XPTY0004', 'argument must contain only nodes')
@@ -1196,7 +1150,7 @@ def select_outermost_function(self: XPathFunction, context: ContextType = None)
             continue
         results.add(item)
 
-    yield from cast(List[XPathNode], sorted(results, key=node_position))
+    yield from cast(list[XPathNode], sorted(results, key=node_position))
 
 
 ##
@@ -1235,7 +1189,7 @@ def evaluate_generate_id_function(self: XPathFunction, context: ContextType = No
 @method(function('uri-collection', nargs=(0, 1),
                  sequence_types=('xs:string?', 'xs:anyURI*')))
 def evaluate_uri_collection_function(self: XPathFunction, context: ContextType = None) \
-        -> List[AnyURI]:
+        -> list[AnyURI]:
     if self.context is not None:
         context = self.context
 
@@ -1268,7 +1222,7 @@ def evaluate_uri_collection_function(self: XPathFunction, context: ContextType =
                 raise self.error('FODC0002', '{!r} collection not found'.format(uri)) from None
 
     if not match_sequence_type(resource_collection, 'xs:anyURI*', self.parser):
-        raise self.error('XPDY0050', "Type does not match sequence type xs:anyURI*")
+        raise self.error('XPDY0050', "type does not match sequence type xs:anyURI*")
 
     return resource_collection
 
@@ -1278,7 +1232,7 @@ def evaluate_uri_collection_function(self: XPathFunction, context: ContextType =
 @method(function('unparsed-text-lines', nargs=(1, 2),
                  sequence_types=('xs:string?', 'xs:string', 'xs:string*')))
 def evaluate_unparsed_text_functions(self: XPathFunction, context: ContextType = None) \
-        -> Union[Emptiable[str], List[str]]:
+        -> Union[Emptiable[str], list[str]]:
     if self.context is not None:
         context = self.context
 
@@ -1416,7 +1370,7 @@ def evaluate_environment_variable_function(self: XPathFunction, context: Context
 @method(function('available-environment-variables', nargs=0,
                  sequence_types=('xs:string*',)))
 def evaluate_available_environment_variables_function(
-        self: XPathFunction, context: ContextType = None) -> List[str]:
+        self: XPathFunction, context: ContextType = None) -> list[str]:
     if self.context is not None:
         context = self.context
     elif context is None:
@@ -1592,7 +1546,7 @@ def evaluate_function_name_function(self: XPathFunction, context: ContextType =
     if not isinstance(func, XPathFunction):
         raise self.error('XPTY0004', "argument is not a function")
     else:
-        name = func.name
+        name = func.qname
         return [] if name is None else name
 
 
@@ -1732,7 +1686,7 @@ def select_namespace_node_kind_test(self: XPathFunction, context: ContextType =
     elif isinstance(context.item, ElementNode):
         elem = context.item
         for context.item in elem.namespace_nodes:
-            yield context.item
+            yield context.item  # noqa
 
 
 ###
@@ -1748,17 +1702,15 @@ XPath30Parser.unregister('round')
 @method(function('data', nargs=(0, 1), sequence_types=('item()*', 'xs:anyAtomicType*')))
 def select_data_function(self: XPathFunction, context: ContextType = None) \
         -> Iterator[AtomicType]:
-    if not self:
-        item = self.get_argument(context, default_to_context=True)
-        value = self.data_value(item)
-        if value is None:
-            raise self.error('FOTY0012', "argument node does not have a typed value")
-        elif isinstance(value, list):
-            yield from value
-        else:
-            yield value
-    else:
+    if self.context is not None:
+        context = self.context
+
+    if self:
         yield from self[0].atomization(context)
+    elif context is None:
+        raise self.missing_context()
+    else:
+        yield from self.atomize_item(context.item)
 
 
 @method(function('document-uri', nargs=(0, 1), sequence_types=('node()?', 'xs:anyURI?')))
@@ -1809,6 +1761,9 @@ def evaluate_node_name_function(self: XPathFunction, context: ContextType = None
         elif name.startswith('{'):
             # name is a QName in extended format
             namespace, local_name = split_expanded_name(name)
+            if not namespace:
+                return QName('', local_name)
+
             for pfx, uri in self.parser.namespaces.items():
                 if uri == namespace:
                     if not pfx:
@@ -1886,9 +1841,7 @@ def evaluate_round_function(self: XPathFunction, context: ContextType = None) \
 # XSD list-based constructors
 
 @XPath30Parser.constructor('NMTOKENS', sequence_types=('xs:NMTOKEN*',))
-def cast_nmtokens_list_type(self: XPathConstructor, value: AtomicType) \
-        -> List[NMToken]:
-    cast_func = xsd10_atomic_types['NMTOKEN']
+def cast_nmtokens_list_type(self: XPathConstructor, value: AtomicType) -> list[NMToken]:
     if isinstance(value, UntypedAtomic):
         values = value.value.split() or [value.value]
     elif hasattr(value, 'split'):
@@ -1897,15 +1850,13 @@ def cast_nmtokens_list_type(self: XPathConstructor, value: AtomicType) \
         raise self.error('FORG0001')
 
     try:
-        return [cast_func(x) for x in values]
+        return [NMToken(x) for x in values]
     except ValueError as err:
         raise self.error('FORG0001', err) from None
 
 
 @XPath30Parser.constructor('IDREFS', sequence_types=('xs:IDREF*',))
-def cast_idrefs_list_type(self: XPathConstructor, value: AtomicType) \
-        -> List[Idref]:
-    cast_func = xsd10_atomic_types['IDREF']
+def cast_idrefs_list_type(self: XPathConstructor, value: AtomicType) -> list[Idref]:
     if isinstance(value, UntypedAtomic):
         values = value.value.split() or [value.value]
     elif hasattr(value, 'split'):
@@ -1914,15 +1865,13 @@ def cast_idrefs_list_type(self: XPathConstructor, value: AtomicType) \
         raise self.error('FORG0001')
 
     try:
-        return [cast_func(x) for x in values]
+        return [Idref(x) for x in values]
     except ValueError as err:
         raise self.error('FORG0001', err) from None
 
 
 @XPath30Parser.constructor('ENTITIES', sequence_types=('xs:ENTITY*',))
-def cast_entities_list_type(self: XPathConstructor, value: AtomicType) \
-        -> List[Entity]:
-    cast_func = xsd10_atomic_types['ENTITY']
+def cast_entities_list_type(self: XPathConstructor, value: AtomicType) -> list[Entity]:
     if isinstance(value, UntypedAtomic):
         values = value.value.split() or [value.value]
     elif hasattr(value, 'split'):
@@ -1931,7 +1880,7 @@ def cast_entities_list_type(self: XPathConstructor, value: AtomicType) \
         raise self.error('FORG0001')
 
     try:
-        return [cast_func(x) for x in values]
+        return [Entity(x) for x in values]
     except ValueError as err:
         raise self.error('FORG0001', err) from None
 
