@@ -9,33 +9,40 @@
 #
 import datetime
 import importlib
+from collections.abc import Iterator, Sequence, Callable
 from copy import copy
+from functools import cached_property
 from types import ModuleType
-from typing import TYPE_CHECKING, cast, Dict, Any, List, Iterator, \
-    Optional, Sequence, Union, Callable, Set, Tuple
+from typing import TYPE_CHECKING, cast, Any, Optional, Union
 
-from .exceptions import ElementPathTypeError
-from .tdop import Token
-from .namespaces import NamespacesType
-from .datatypes import AnyAtomicType, Timezone, Language
-from .protocols import ElementProtocol, DocumentProtocol
-from .etree import is_etree_element, is_etree_document
-from .xpath_nodes import ChildNodeType, XPathNode, AttributeNode, NamespaceNode, \
-    CommentNode, ProcessingInstructionNode, ElementNode, DocumentNode, SchemaElementNode
-from .tree_builders import RootArgType, get_node_tree
+from elementpath.aliases import NamespacesType, SequenceType, InputType
+from elementpath.protocols import ElementProtocol, DocumentProtocol
+from elementpath.exceptions import ElementPathTypeError
+from elementpath.tdop import Token
+from elementpath.datatypes import AnyAtomicType, AtomicType, Timezone, Language
+from elementpath.etree import is_etree_element, is_etree_element_instance, is_etree_document
+from elementpath.xpath_nodes import ChildNodeType, XPathNode, AttributeNode, NamespaceNode, \
+    CommentNode, ProcessingInstructionNode, ElementNode, DocumentNode, RootNodeType, \
+    RootArgType
+from elementpath.tree_builders import get_node_tree
 
 if TYPE_CHECKING:
-    from .xpath_tokens import XPathToken, XPathAxis, XPathFunction
-    ItemType = Union[None, XPathNode, AnyAtomicType, XPathFunction]
-    ItemArgType = Union[XPathNode, AnyAtomicType, XPathFunction,
-                        ElementProtocol, DocumentProtocol]
-else:
-    ItemType = Any
-    ItemArgType = Any
+    from elementpath.schema_proxy import AbstractSchemaProxy
+    from elementpath.xpath_tokens import XPathToken, XPathAxis, XPathFunction  # noqa
 
-__all__ = ['XPathContext', 'XPathSchemaContext']
+__all__ = ['XPathContext', 'XPathSchemaContext', 'ContextType', 'ItemType',
+           'ValueType', 'ItemArgType', 'FunctionArgType']
 
-CollectionArgType = Union[None, ItemArgType, List[ItemArgType], Tuple[ItemArgType, ...]]
+###
+# type annotations aliases for context and tokens classes
+ContextType = Union['XPathContext', 'XPathSchemaContext', None]
+ItemType = Union[XPathNode, AtomicType, 'XPathFunction']
+ValueType = SequenceType[ItemType]
+ItemArgType = Union[ItemType, ElementProtocol, DocumentProtocol]
+FunctionArgType = Union[InputType[ItemArgType], ValueType]
+
+NodeArgType = Union[XPathNode, ElementProtocol, DocumentProtocol]
+CollectionArgType = Optional[InputType[NodeArgType]]
 
 
 class XPathContext:
@@ -56,15 +63,17 @@ class XPathContext:
     This can be useful when the dynamic context has additional namespaces and root \
     is an Element or an ElementTree instance of the standard library.
     :param uri: an optional URI associated with the root element or the document.
-    :param fragment: if `True` a root element is considered a fragment, otherwise \
-    a root element is considered the root of an XML document, and a dummy document \
-    is created for selection. In this case the dummy document value is not included \
-    in the results.
+    :param fragment: if `True` is provided the root is considered a fragment. In this \
+    case if `root` is an ElementTree instance skips it and use the root Element. If \
+    `False` is provided creates a dummy document when the root is an Element instance. \
+    In this case the dummy document value is not included in results. For default the \
+    root node kind is preserved.
     :param item: the context item. A `None` value means that the context is positioned on \
     the document node.
     :param position: the current position of the node within the input sequence.
     :param size: the number of items in the input sequence.
     :param axis: the active axis. Used to choose when apply the default axis ('child' axis).
+    :param schema: an optional schema proxy instance to be applied on XDM root or item.
     :param variables: dictionary of context variables that maps a QName to a value.
     :param current_dt: current dateTime of the implementation, including explicit timezone.
     :param timezone: implicit timezone to be used when a date, time, or dateTime value does \
@@ -86,40 +95,47 @@ class XPathContext:
     and fn:available-environment-variables.
     """
     _etree: Optional[ModuleType] = None
-    root: Union[DocumentNode, ElementNode, None] = None
-    document: Optional[DocumentNode] = None
-    item: Optional[ItemType]
-    total_nodes: int = 0  # Number of nodes associated to the context
+    _schema: Optional['AbstractSchemaProxy'] = None
+    root: Optional[RootNodeType]
+    document: Optional[DocumentNode]
+    item: ItemType
 
-    variables: Dict[str, Union[ItemType, List[ItemType]]]
-    documents: Optional[Dict[str, Union[DocumentNode, ElementNode, None]]] = None
-    collections = None
-    default_collection = None
+    variables: dict[str, ValueType]
+    documents: Optional[dict[str, DocumentNode]] = None
+    collections: Optional[dict[str, list[XPathNode]]] = None
+    default_collection: Optional[list[XPathNode]] = None
+
+    __slots__ = ('document', 'root', 'item', 'namespaces', 'size',
+                 'position', 'variables', 'axis', '__dict__')
 
     def __init__(self,
                  root: Optional[RootArgType] = None,
                  namespaces: Optional[NamespacesType] = None,
                  uri: Optional[str] = None,
-                 fragment: bool = False,
+                 fragment: Optional[bool] = None,
                  item: Optional[ItemArgType] = None,
                  position: int = 1,
                  size: int = 1,
                  axis: Optional[str] = None,
-                 variables: Optional[Dict[str, CollectionArgType]] = None,
+                 schema: Optional['AbstractSchemaProxy'] = None,
+                 variables: Optional[dict[str, InputType[ItemArgType]]] = None,
                  current_dt: Optional[datetime.datetime] = None,
                  timezone: Optional[Union[str, Timezone]] = None,
-                 documents: Optional[Dict[str, Optional[RootArgType]]] = None,
-                 collections: Optional[Dict[str, CollectionArgType]] = None,
+                 documents: Optional[dict[str, RootArgType]] = None,
+                 collections: Optional[dict[str, CollectionArgType]] = None,
                  default_collection: CollectionArgType = None,
-                 text_resources: Optional[Dict[str, str]] = None,
-                 resource_collections: Optional[Dict[str, List[str]]] = None,
+                 text_resources: Optional[dict[str, str]] = None,
+                 resource_collections: Optional[dict[str, list[str]]] = None,
                  default_resource_collection: Optional[str] = None,
                  allow_environment: bool = False,
                  default_language: Optional[str] = None,
                  default_calendar: Optional[str] = None,
                  default_place: Optional[str] = None) -> None:
 
-        self.namespaces = dict(namespaces) if namespaces else {}
+        if namespaces:
+            self.namespaces = {k: v for k, v in namespaces.items()}
+        else:
+            self.namespaces = {}
 
         if root is not None:
             self.root = get_node_tree(root, self.namespaces, uri, fragment)
@@ -129,21 +145,20 @@ class XPathContext:
                 self.item = self.root
 
         elif item is not None:
+            self.root = None
             self.item = self.get_context_item(item, self.namespaces, uri, fragment)
         else:
             raise ElementPathTypeError("Missing both the root node and the context item!")
 
         if isinstance(self.root, DocumentNode):
             self.document = self.root
-        elif fragment or isinstance(self.root, SchemaElementNode):
-            pass
-        elif isinstance(self.root, ElementNode):
-            # Creates a dummy document
-            document = cast(DocumentProtocol, self.etree.ElementTree(self.root.elem))
-            self.document = DocumentNode(document, self.root.uri, position=0)
-            self.document.children.append(self.root)
-            self.document.elements = cast(Dict[ElementProtocol, ElementNode], self.root.elements)
-            # self.root.parent = self.document  TODO for v5? It's necessary?
+        elif fragment is None and \
+                isinstance(self.root, ElementNode) and \
+                is_etree_element_instance(self.root.obj):
+            # Creates a dummy document that will be not included in results
+            self.document = self.root.get_document_node(as_parent=False)
+        else:
+            self.document = None
 
         self.position = position
         self.size = size
@@ -156,18 +171,19 @@ class XPathContext:
         self.current_dt = current_dt or datetime.datetime.now(tz=self.timezone)
 
         if documents is not None:
+            # Assume that are all documents because type checking is done by fn:doc().
             self.documents = {
-                k: get_node_tree(v, self.namespaces, k) if v is not None else v
-                for k, v in documents.items()
+                k: cast(DocumentNode, get_node_tree(v, self.namespaces, k))
+                if v is not None else v for k, v in documents.items()
             }
 
+        if schema is not None:
+            self.schema = schema
+
         self.variables = {}
         if variables is not None:
-            for k, v in variables.items():
-                if v is None or isinstance(v, (list, tuple)):
-                    self.variables[k] = self.get_collection(v)
-                else:
-                    self.variables[k] = self.get_context_item(v)
+            for varname, value in variables.items():
+                self.variables[varname] = self.get_value(value, self.namespaces)
 
         if collections is not None:
             self.collections = {k: self.get_collection(v) for k, v in collections.items()}
@@ -185,47 +201,71 @@ class XPathContext:
 
     def __repr__(self) -> str:
         if self.root is not None:
-            return f'{self.__class__.__name__}(root={self.root.value})'
+            return f'{self.__class__.__name__}(root={self.root.obj})'
         elif isinstance(self.item, XPathNode):
-            return f'{self.__class__.__name__}(item={self.item.value})'
+            return f'{self.__class__.__name__}(item={self.item.obj})'
         else:
             return f'{self.__class__.__name__}(item={self.item!r})'
 
     def __copy__(self) -> 'XPathContext':
         obj: XPathContext = object.__new__(self.__class__)
         obj.__dict__.update(self.__dict__)
+        obj.document = self.document
+        obj.root = self.root
+        obj.item = self.item
+        obj.size = self.size
+        obj.position = self.position
         obj.axis = None
+        obj.namespaces = {k: v for k, v in self.namespaces.items()}
         obj.variables = {k: v for k, v in self.variables.items()}
         return obj
 
-    @property
+    @cached_property
     def etree(self) -> ModuleType:
-        if self._etree is None:
-            if isinstance(self.root, (DocumentNode, ElementNode)):
-                module_name = self.root.value.__class__.__module__
-            elif isinstance(self.item, (DocumentNode, ElementNode, CommentNode,
-                                        ProcessingInstructionNode)):
-                module_name = self.item.value.__class__.__module__
-            else:
-                module_name = 'xml.etree.ElementTree'
-
-            if not isinstance(module_name, str) or not module_name.startswith('lxml.'):
-                etree_module_name = 'xml.etree.ElementTree'
-            else:
-                etree_module_name = 'lxml.etree'
-
-            self._etree: ModuleType = importlib.import_module(etree_module_name)
-
-        return self._etree
-
-    def get_root(self, node: Any) -> Union[None, ElementNode, DocumentNode]:
         if isinstance(self.root, (DocumentNode, ElementNode)):
-            if any(node is x for x in self.root.iter()):
+            module_name = self.root.obj.__class__.__module__
+        elif isinstance(self.item, (DocumentNode, ElementNode, CommentNode,
+                                    ProcessingInstructionNode)):
+            module_name = self.item.obj.__class__.__module__
+        else:
+            module_name = 'xml.etree.ElementTree'
+
+        if module_name in ('lxml.etree', 'lxml.html'):
+            return importlib.import_module('lxml.etree')
+        else:
+            return importlib.import_module('xml.etree.ElementTree')
+
+    @property
+    def schema(self) -> Optional['AbstractSchemaProxy']:
+        return self._schema
+
+    @schema.setter
+    def schema(self, schema: Optional['AbstractSchemaProxy']) -> None:
+        self._schema = schema
+        if schema is None:
+            if self.root is not None:
+                self.root.clear_types()
+            elif isinstance(self.item, XPathNode):
+                self.item.clear_types()
+        elif hasattr(schema, 'is_assertion_based'):
+            if self.root is not None:
+                self.root.clear_types()
+                self.root.apply_schema(schema)
+            elif isinstance(self.item, XPathNode):
+                self.item.clear_types()
+                self.item.apply_schema(schema)
+        else:
+            msg = f"{schema!r} is not an instance of AbstractSchemaProxy"
+            raise ElementPathTypeError(msg)
+
+    def get_root(self, node: Any) -> Optional[RootNodeType]:
+        if isinstance(self.root, (DocumentNode, ElementNode)):
+            if any(node is x for x in self.root.iter_lazy()):
                 return self.root
 
         if self.documents is not None:
             for uri, doc in self.documents.items():
-                if doc is not None and any(node is x for x in doc.iter()):
+                if doc is not None and any(node is x for x in doc.iter_lazy()):
                     return doc
 
         return None
@@ -250,7 +290,7 @@ class XPathContext:
     def get_context_item(self, item: ItemArgType,
                          namespaces: Optional[NamespacesType] = None,
                          uri: Optional[str] = None,
-                         fragment: bool = False) -> ItemType:
+                         fragment: Optional[bool] = None) -> ItemType:
         """
         Checks the item and returns an item suitable for XPath processing.
         For XML trees and elements try a match with an existing node in the
@@ -260,12 +300,12 @@ class XPathContext:
         if isinstance(item, (XPathNode, AnyAtomicType)):
             return item
         elif is_etree_document(item):
-            if self.root is not None and item is self.root.value:
+            if self.root is not None and item is self.root.obj:
                 return self.root
 
             if self.documents:
                 for doc in self.documents.values():
-                    if doc is not None and item is doc.value:
+                    if doc is not None and item is doc.obj:
                         return doc
 
         elif is_etree_element(item):
@@ -277,7 +317,7 @@ class XPathContext:
             if self.documents:
                 for doc in self.documents.values():
                     if doc is not None and doc.elements is not None and item in doc.elements:
-                        return doc.elements[item]  # type: ignore[index]
+                        return doc.elements[item]
 
             if callable(item.tag):  # type: ignore[union-attr]
                 if item.tag.__name__ == 'Comment':  # type: ignore[union-attr]
@@ -297,15 +337,24 @@ class XPathContext:
             fragment=fragment
         )
 
-    def get_collection(self, items: Optional[CollectionArgType]) -> Optional[List[ItemType]]:
-        if items is None:
-            return None
-        elif isinstance(items, (list, tuple)):
-            return [self.get_context_item(x) for x in items]
-        else:
-            return [self.get_context_item(items)]
+    def get_value(self, item: FunctionArgType, *args: Any, **kwargs: Any) -> ValueType:
+        if item is None:
+            return []
+        elif not isinstance(item, (list, tuple)):
+            return self.get_context_item(item, *args, **kwargs)
+        return [self.get_context_item(x, *args, **kwargs) for x in item]
 
-    def inner_focus_select(self, token: Union['XPathToken', 'XPathAxis']) -> Iterator[Any]:
+    def get_collection(self, items: CollectionArgType) -> list[XPathNode]:
+        if items is None:
+            return []
+        elif isinstance(items, (list, tuple)):
+            return [x for x in map(self.get_context_item, items) if isinstance(x, XPathNode)]
+        else:
+            item = self.get_context_item(items)
+            return [item] if isinstance(item, XPathNode) else []
+
+    def inner_focus_select(self, token: Union['XPathToken', 'XPathAxis'], predicate: bool = False) \
+            -> Iterator[ItemType]:
         """Apply the token's selector with an inner focus."""
         status = self.item, self.size, self.position, self.axis
         results = [x for x in token.select(copy(self))]
@@ -331,6 +380,8 @@ class XPathContext:
         :param selectors: a sequence of selector generator functions.
         :param varnames: a sequence of variables for storing the generated values.
         """
+        if varnames is None:
+            varnames = []
         iterators = [x(self) for x in selectors]
         dimension = len(iterators)
         prod = [None] * dimension
@@ -338,30 +389,28 @@ class XPathContext:
 
         k = 0
         while True:
-            try:
-                value = next(iterators[k])
-            except StopIteration:
-                if not k:
-                    return
-                iterators[k] = selectors[k](self)
-                k -= 1
-            else:
-                if varnames is not None:
-                    try:
-                        self.variables[varnames[k]] = value
-                    except (TypeError, IndexError):
-                        pass
+            for value in iterators[k]:
+                try:
+                    self.variables[varnames[k]] = value
+                except IndexError:
+                    pass
 
                 prod[k] = value
                 if k == max_index:
                     yield tuple(prod)
                 else:
                     k += 1
+                break
+            else:
+                if not k:
+                    return
+                iterators[k] = selectors[k](self)
+                k -= 1
 
     ##
     # Context item iterators for axis
 
-    def iter_self(self) -> Iterator[Optional[ItemType]]:
+    def iter_self(self) -> Iterator[ItemType]:
         """Iterator for 'self' axis and '.' shortcut."""
         if self.item is not None:
             status = self.axis
@@ -388,7 +437,7 @@ class XPathContext:
 
             self.item, self.axis = status
 
-    def iter_children_or_self(self) -> Iterator[Optional[ItemType]]:
+    def iter_children_or_self(self) -> Iterator[ItemType]:
         """Iterator for 'child' forward axis and '/' step."""
         if self.item is not None:
             if self.axis is not None:
@@ -398,14 +447,40 @@ class XPathContext:
                 self.axis = 'child'
 
                 if self.item is self.document and self.root is not self.document:
-                    yield self.root
+                    if self.root is not None:
+                        yield self.root
                 else:
                     for self.item in self.item:
                         yield self.item
 
                 self.item, self.axis = _status
 
-    def iter_parent(self) -> Iterator[Union[ElementNode, DocumentNode]]:
+    def iter_matching_nodes(self, name: str, default_namespace: Optional[str] = None) \
+            -> Iterator[Union[AttributeNode, ElementNode]]:
+        """
+        Iterator for matching elements or attributes. For default uses 'child'
+        forward axis if no axis is active, otherwise tests the current item.
+        """
+        if self.axis is not None:
+            if isinstance(self.item, (AttributeNode, ElementNode)):
+                if self.item.match_name(name, default_namespace):
+                    yield self.item
+        elif isinstance(self.item, (ElementNode, DocumentNode)):
+            _status = self.item, self.axis
+            self.axis = 'child'
+
+            if self.item is self.document and isinstance(self.root, ElementNode):
+                if self.root.match_name(name, default_namespace):
+                    yield self.root
+            else:
+                for self.item in self.item:
+                    if self.item.match_name(name, default_namespace):
+                        assert isinstance(self.item, ElementNode)
+                        yield self.item
+
+            self.item, self.axis = _status
+
+    def iter_parent(self) -> Iterator[RootNodeType]:
         """Iterator for 'parent' reverse axis and '..' shortcut."""
         if isinstance(self.item, XPathNode):
 
@@ -481,7 +556,7 @@ class XPathContext:
             status = self.item, self.axis
             self.axis = axis or 'ancestor'
 
-            ancestors: List[XPathNode] = []
+            ancestors: list[XPathNode] = []
             if axis == 'ancestor-or-self':
                 ancestors.append(self.item)
 
@@ -500,7 +575,7 @@ class XPathContext:
 
     def iter_preceding(self) -> Iterator[Union[DocumentNode, ChildNodeType]]:
         """Iterator for 'preceding' reverse axis."""
-        ancestors: Set[Union[ElementNode, DocumentNode]]
+        ancestors: set[RootNodeType]
         item: XPathNode
 
         if isinstance(self.item, XPathNode):
@@ -542,7 +617,7 @@ class XPathContext:
             for item in root.iter_descendants(with_self=False):
                 if position < item.position and item not in descendants:
                     self.item = item
-                    yield cast(ChildNodeType, self.item)
+                    yield item
 
             self.item, self.axis = status
 
@@ -555,3 +630,52 @@ class XPathSchemaContext(XPathContext):
     XML instances.
     """
     root: ElementNode
+
+    @property
+    def schema(self) -> Optional['AbstractSchemaProxy']:
+        return self._schema
+
+    @schema.setter
+    def schema(self, schema: Optional['AbstractSchemaProxy']) -> None:
+        self._schema = schema
+
+    def iter_matching_nodes(self, name: str, default_namespace: Optional[str] = None) \
+            -> Iterator[Union[AttributeNode, ElementNode]]:
+        """
+        Iterator for matching elements or attributes. For default uses 'child'
+        forward axis if no axis is active, otherwise tests the current item.
+        """
+        if self.axis is not None:
+            if isinstance(self.item, (AttributeNode, ElementNode)):
+                if self.item.match_name(name, default_namespace):
+                    if not self.item.name:
+                        if isinstance(self.item, ElementNode):
+                            for element_node in self.root:
+                                assert isinstance(element_node, ElementNode)
+                                if element_node.match_name(name, default_namespace):
+                                    self.item = element_node
+                                    break
+                        else:
+                            for attribute_node in self.root.attributes:
+                                if attribute_node.match_name(name, default_namespace):
+                                    self.item = attribute_node
+                                    break
+
+                    yield self.item
+
+        elif isinstance(self.item, ElementNode):
+            _status = self.item, self.axis
+            self.axis = 'child'
+
+            for self.item in self.item:
+                if self.item.match_name(name, default_namespace):
+                    if not self.item.name:
+                        for element_node in self.root:
+                            if element_node.match_name(name, default_namespace):
+                                self.item = element_node
+                                break
+
+                    assert isinstance(self.item, ElementNode)
+                    yield self.item
+
+            self.item, self.axis = _status
